package algorithmrepository.differentialgeometry;

import algorithmrepository.interfaces.ScalarND;
import algorithmrepository.interfaces.VectorND;

/**
 * Defines a general curvilinear geometry based on a metric tensor and a forward coordinate
 * transformation from the coordinates of the geometry (u1, u2, u3..), to Cartesian coordinates (x,y,z..). 
 * 
 * It handles and lets the user use either the covariant or contravariant components of vectors defined in the 
 * geometry. 
 * 
 * It can also calculate differential elements in the geometry (dl, dS, dV), and derivative operators over fields
 * defined in the geometry (for example div, grad, curl), and dot and cross products in the covariant or contravariant basis.
 * 
 * See Chapter 2 in "Flux Coordinate and Magnetic Field Structure", WD D'haeseleer et al for extremely good description of 
 * all what is implemented here.
 * 
 * @author Jakob Svensson (c) 2014
 *
 */
public interface Geometry {
    
    public static final int COVARIANT = 0;
    public static final int CONTRAVARIANT = 1;
        
    /**
     * Returns the metric tensor for the geometry, in either its covariant or contravariant form.
     * 
     * @param atCoords Coordinates for the position at which the metric should be evaluated. 
     * @param coOrContraMetric Either COVARIANT or CONTRAVARIANT
     * 
     * @return The metric tensor at the given coordinates.
     */
    public double[][] metricTensor(double[] atCoords, int coOrContraMetric);   
    
    
    /**
     * Calculates a covariant or contravariant metric coefficient from central difference derivatives of the position vectors w.r.t the geometry coordinates.
     * The step length used are given/set by getStepLengthCoordDerivatives/setStepLengthCoordDerivatives.
     * 
     * @param atCoords The coordinates (in the geometry) where the metric coefficent should be evaluated.
     * @param i The row index of the coefficient in the metric tensor.
     * @param j The column index of the coefficient in the metric tensor.
     * @param coOrContraVariant Whether the COVARIANT or CONTRAVARIANT metric coefficient should be calculated.
     * @return The metric coefficient gij.
     */
    public double metricCoefficientFromCoordinateTransform(double[] atCoords, int i, int j, int coOrContraVariant);
    
    /**
     * Calculates the covariant or contravariant metric tensor from central difference derivatives of the position vectors w.r.t the geometry coordinates.
     * The step length used are given/set by getStepLengthCoordDerivatives/setStepLengthCoordDerivatives.
     * 
     * @param atCoords The coordinates (in the geometry) where the metric tensor should be evaluated.
     * @param coOrContraVariant Whether the COVARIANT or CONTRAVARIANT metric tenor should be calculated.
     * @return The metric tensor
     */    
    public double[][] metricTensorFromCoordinateTransform(double[] atCoords, int coOrContraVariant);
    
    /**
     * Returns an individual metric coefficient from the metric tensor, in either its covariant or contravariant form.
     * 
     * @param atCoords Coordinates for the position at which the metric should be evaluated. 
     * @param coOrContraMetric Either COVARIANT or CONTRAVARIANT
     * 
     * @return The given metric coefficient at the given coordinates.
     */
    public double metricCoefficient(double[] atCoords, int i, int j, int coOrContra);
    
    /**
     * Derivative of a metric coefficient with respect to a coordinate (dgij/duk). 
     * The default implementation uses numerical central differences
     * with step length in each coordinate direction given/set by getStepLengthCoordDerivatives/setStepLengthCoordDerivatives.
     * 
     * @param atCoords The coordinates at which the metric derivative should be calculated.
     * @param i Row index in metric tensor
     * @param j Column index in metric tensor
     * @param k Coordinate index for the coordinate the derivative is with respect to.
     * @param coOrContra Determines whether the COVARIANT or CONTRAVARIANT metric should be used.  
     * @return The derivative dgij/duk.
     */
    public double metricCoeffDerivative(double[] atCoords, int i, int j, int k, int coOrContra);
    
    /**
     * Transforms from COVARIANT to CONTRAVARIANT form of a vectors components.
     * 
     * @param atCoords Position at which the transformation should be done.
     * @param coComponents The components in COVARIATE form.
     * @return The components in CONTRAVARIANT form.
     */
    public double[] coToContra(double[] atCoords, double[] coComponents);
    
    /**
     * Transforms from CONTRAVARIANT to COVARIANT form of a vectors components.
     * 
     * @param atCoords Position at which the transformation should be done.
     * @param contraComponents The components in CONTRAVARIANT form.
     * @return The components in COVARIANT form.
     */    
    public double[] contraToCo(double[] atCoords, double[] contraComponents);
    
    /**
     * Transforms from CONTRAVARIANT to COVARIANT and vice versa.
     * 
     * @param atCoords Position at which the transformation should be done.
     * @param components The components in CONTRAVARIANT form.
     * @param componentsCoContra The format of the given component vector, either COVARIANT
     * or CONTRAVARIANT.
     * @return The components flipped to the opposite basis.
     */        
    public double[] flipCoContra(double[] atCoords, double[] components, int componentsCoContra);
    
    /**
     * The Jacobian determinant of x,y,z w.r.t. u1,u2,u3, which is the same as sqrt(g) (g being
     * the determinant of the metric tensor).
     * 
     * @param atCoords The coordinates at which the Jacobian determinant should be calculated.
     * @return The Jacobian determinant at the given position.
     */
    public double jacobianDeterminant(double[] atCoords);
    
    
    /**
     * Jacobian determinants of x,y,z w.r.t. u1,u2,u3 at multiple points. The Jacobian determinant 
     * is the same as sqrt(g), g being  the determinant of the metric tensor.
     * 
     * @param atCoords The coordinates at which the Jacobian determinant should be calculated.
     * atCoords[i][j] is the i:th coordinate for position number j.
     * @return The Jacobian determinants at the given positions.
     */
    public double[] jacobianDeterminant(double[][] atCoords);
    
        
    /**
     * Returns the Jacobian matrix of x,y,z w.r.t. u1,u2,u3.
     * 
     * @param atCoords The position (u1,u2,u3) where the Jacobian should be evaluated.
     * @return The Jacobian matrix of x,y,z w.r.t. u1,u2,u3.
     */
    public double[][] jacobian(double[] atCoords);
    
    /**
     * Returns the i,j:th element of the Jacobian matrix, so dv(i)/du(j) with i==0 for x, 
     * i == 1 for y, and i == 2 for z.
     * 
     * @param atCoords The position (u1,u2,u3) where the Jacobian matrix element should be evaluated.
     * @return The i,jth Jacobian matrix element of x,y,z w.r.t. u1,u2,u3.
     */
    public double jacobianElement(double[] atCoords, int i, int j);

    /**
     * Dot product of two vectors, both represented in either COVARIANT or CONTRAVARIANT components.
     * 
     * @param atCoords The position at which the dot product should be evaluted.
     * @param v1 First vector of the dot product
     * @param v2 Second vector of the dot product.
     * @param coOrContra Component form of the vectors, either COVARIANT or CONTRAVARIANT.
     * 
     * @return The dot product of the given vectors.
     */
    public double dot(double[] atCoords, double[] v1, double[] v2, int coOrContra);
    
    /**
     * Cross product of two vectors, both represented in either COVARIANT or CONTRAVARIANT components.
     * 
     * @param atCoords The position at which the cross product should be evaluated.
     * @param v1 First vector of the cross product
     * @param v2 Second vector of the cross product.
     * @param coOrContra Component form of the vectors, either COVARIANT or CONTRAVARIANT.
     * 
     * @return The cross product of the given vectors. The result is in the same component form
     * as the given input vectors. 
     */    
    public double[] cross(double[] atCoords, double[] v1, double[] v2, int coOrContra);

    /**
     * Differential element dl along one coordinate line.
     * 
     * @param atCoords The position at which the coordinate differential element should be calculated.
     * @param dim The index of the coordinate whose differential should be calculated.
     * @param du The unscaled differential element. 
     * @return The (scaled) differential element, to be used in integrals along one coordinate direction.
     */
    public double[] dl(double[][] atCoords, int dim, double du);
    
    
    /**
     * Differential surface element dS in the surface of two coordinates, keeping the third constant.
     * 
     * @param atCoords The position at which the coordinate differential element should be calculated.
     * @param dim1 The index of the first coordinate direction
     * @param dim2 The index of the second coordinate direction
     * @param du1 The unscaled differential element in first coordinate direction.
     * @param du2 The unscaled differential element in second coordinate direction.
     * @return The (scaled) differential element, to be used in integrals over a surface spanned by the two given
     * coordinates.
     */    
    public double[] dS(double[][] atCoords, int dim1, int dim2, double du1, double du2);

    /**
     * Differential volume element dV of the geometry.
     * 
     * @param atCoords The position at which the coordinate differential volume element should be calculated.
     * @param du Array of the unscaled differential elements in each direction.
     * @return The (scaled) differential volume element, to be used in integrals over volumes in the geometry.
     */    
    public double[] dV(double[][] atCoords, double[] du);   
   
    /**
     * Calculates the Christoffel symbol of the second kind, which is the upperDim contravariant
     * component of the de/duk, where e is the lowerDim1 covariant basis vector, and uk the lowerDim2
     * coordinate. The naming follows the visual representation of the Chritoffel symbol.
     * 
     * @param upperDim Upper dimension in the Christoffel symbol (see above).
     * @param lowerDim1 Left lower dimension in Christoffel symbol (see above).
     * @param lowerDim2 Right lower dimension in Christoffel symbol (see above).
     * @return The Christoffer symbol of the second kind at the given position (see above).
     */
    public double christoffelSymbol2ndKind(double[] atCoords, int upperDim, int lowerDim1, int lowerDim2);
    
    /**
     * Calculates the Christoffel symbol of the first kind, which is the upperDim covariant
     * component of the de/duk, where e is the lowerDim1 covariant basis vector, and uk the lowerDim2
     * coordinate. The naming follows the visual representation of the Chritoffel symbol.
     * 
     * @param upperDim Upper dimension in the Christoffel symbol (see above).
     * @param lowerDim1 Left lower dimension in Christoffel symbol (see above).
     * @param lowerDim2 Right lower dimension in Christoffel symbol (see above).
     * @return The Christoffer symbol of the first kind at the given position (see above).
     */    
    public double christoffelSymbol1stKind(double[] atCoords, int upperDim, int lowerDim1, int lowerDim2);
            
    /**
     * Sets the step length for all coordinate directions. These step lengths are used for all 
     * default derivative calculations. 
     * 
     * @param du The step lengths used by the Geometry class, for all default derivative calculations with 
     * respect to coordinates.
     */
    public void setStepLengthCoordDerivatives(double[] du);
    
    /**
     * Gets the step length for all coordinate directions. These step lengths are used for all 
     * default derivative calculations. 
     * 
     * @return The step lengths used by the Geometry class, for all default derivative calculations with 
     * respect to coordinates.
     */    
    public double[] getStepLengthCoordDerivatives();    
    
    /**
     * Evaluates a given vector field at multiple positions. The vector field domain is the coordinates of this geometry 
     * (u1, u2, u3..), and the fields range are vectors in either COVARIANT or CONTRAVARIANT form, which must be specified
     * in the call.
     *  
     * @param field The field to evaluate. Must have the coordinates of the geometry as domain, and vectors in either COVARIANT
     * or CONTRAVARIANT form as range.
     * @param fieldCoOrContra The form of the vector fields range (COVARIANT or CONTRAVARIANT)
     * @param returnCoOrContra The form of the returned vectos (COVARIANT or CONTRAVARIANT). Can be different
     * from the fields form.
     * @param atCoords The positions at which the vector field should be evaluated. atCoords[i][j] is the i:th coordinate 
     * for position number j.
     * @return The vectors at the given positions.
     */
    public double[][] evalVectorField(VectorND field, int fieldCoOrContra, int returnCoOrContra, double[][] atCoords);
    
    /**
     * Evaluates a given vector field at a single position. The vector field domain is the coordinates of this geometry 
     * (u1, u2, u3..), and the fields range are vectors in either COVARIANT or CONTRAVARIANT form, which must be specified
     * in the call.
     *  
     * @param field The field to evaluate. Must have the coordinates of the geometry as domain, and vectors in either COVARIANT
     * or CONTRAVARIANT form as range.
     * @param fieldCoOrContra The form of the vector fields range (COVARIANT or CONTRAVARIANT)
     * @param returnCoOrContra The form of the returned vectos (COVARIANT or CONTRAVARIANT). Can be different
     * from the fields form.
     * @param atCoords The position at which the vector field should be evaluated. atCoords[i] is the i:th coordinate value.
     * @return The vectors at the given position.
     */    
    public double[] evalVectorField(VectorND field, int fieldCoOrContra, int returnCoOrContra, double[] atCoords);
    
    /**
     * Calculates the divergence of a vector field. The numerical differentials are carried out using
     * a central difference calculation with step lengths given/set by getStepLengthCoordDerivatives/setStepLengthCoordDerivatives.
     * 
     * @param field The field whose divergence should be calculated.
     * @param fieldCoOrContra Whether the vector field has COVARIANT or CONTRAVARIANT components.
     * @param atCoords The position (coordinates of the geometry) at which the divergence should be calculated.
     * @return The divergence of the vector field at the given position.
     */
    public double div(VectorND field, int fieldCoOrContra, double[] atCoords); 

    /**
     * Calculates the gradient of a scalar field. The numerical differentials are carried out using
     * a central difference calculation with step lengths given/set by getStepLengthCoordDerivatives/setStepLengthCoordDerivatives.
     * 
     * @param field The scalar field whose gradient should be calculated.
     * @param returnCoOrContra Whether the gradient vector field should have COVARIANT or CONTRAVARIANT components.
     * @param atCoords The position (coordinates of the geometry) at which the gradient should be calculated.
     * @return The gradient of the field at the given position.
     */
    public double[] grad(ScalarND field, int returnCoOrContra, double[] atCoords);
    
    /**
     * Calculates the curl of a vector field. The numerical differentials are carried out using
     * a central difference calculation with step lengths given/set by getStepLengthCoordDerivatives/setStepLengthCoordDerivatives.
     * 
     * @param field The vector field whose curl should be calculated.
     * @param fieldCoOrContra Whether the given field has COVARIANT or CONTRAVARIANT components.
     * @param returnCoOrContra Whether the returned vector should have COVARIANT or CONTRAVARIANT components.
     * @param atCoords The position (coordinates of the geometry) at which the curl should be calculated.
     * @return The curl of the vector field at the given position.
     */    
    public double[] curl(VectorND field, int fieldCoOrContra, int returnCoOrContra, double[] atCoords);
        
    /**
     * Calculates the COVARIANT or CONTRAVARIANT basis vectors at a given position, and returns them
     * in Cartesian components.
     * 
     * @param atCoords The coordinates (u1, u2, u3..) at which the basis vectors should be calculated.
     * @param coOrContraBasisVectors Whether the COVARIANT or CONTRAVARIANT basis vectors should be calculated. 
     * @return The basis vectors at a given position, in Cartesian components, so ret[i][j] is i:th basis vector, and
     * j=0,1,2 for the x,y,z components.
     */
    public double[][] cartesianBasisVectors(double[] atCoords, int coOrContraBasisVectors);
        
    /**
     * Transforms a vector with components in either COVARIANT or CONTRAVARIANT form, to Cartesian (x,y,z..) components.
     * 
     * @param atCoords The coordinates (u1, u2 ,u3..) at which the vector is positioned.
     * @param components The components of the vector.
     * @param componentsCoOrContra Whether the vector components are in COVARIANT or CONTRAVARIANT form.
     * 
     * @return The components of the vector along the Cartesian axis.
     */
    public double[] toCartesianComponents(double[] atCoords, double[] components, int componentsCoOrContra);
    
    /**
     * Transforms a Cartesian  vector at given geometry coordinates, to either COVARIANT or CONTRAVARIANT form.
     * 
     * @param atCoords The coordinates (u1, u2 ,u3..) at which the vector is positioned.
     * @param componentsXYZ The Cartesian (x,y,z..) components of the vector.
     * @param componentsCoOrContra Whether the returned vector components should be in COVARIANT or CONTRAVARIANT form.
     * 
     * @return The COVARIANT or CONTRAVARIANT components of a Cartesian vector at the given coordinates.
     */    
    public double[] fromCartesianComponents(double[] atCoords, double[] componentsXYZ, int returnCoOrContra); 
    
    /**
     * The dimensionality of the geometry. Has to be implemented by an implementing class.
     * 
     * @return The dimensionality of this geometry.
     */
    public int dim();

    /**
     * The forward transformation from geometry coordinates (u1, u2, u3...) to Cartesian coordinates (x,y,z..).
     * Has to be implemented by an implementing class.
     * 
     * @param coords Geometry coordinates (u1, u2, u3..) to be transformed to Cartesian coordinates. 
     * @return The Cartesian coordinates corresponding to the given Geometry coordinates.
     */    
    public double[] toCartesianCoords(double[] coords);
    
    /**
     * This method is not used by any other method in the framework, and its default implementation
     * throws a NotImplementedException. It is only defined in the interface to allow the user to implement
     * this back-transformation if wished for, since the forward transformation (toCartesianCoords) does 
     * anyway have to be implemented.
     * 
     * @param coordsXYZ The cartesian coordinates (x,y,z..).
     * @return The coordinates (u1, u2, u3..) defined by this class.
     */
    public double[] fromCartesianCoords(double[] coordsXYZ);
    
}
