package algorithmrepository.differentialgeometry;

import algorithmrepository.Algorithms;
import algorithmrepository.exceptions.NotImplementedException;
import algorithmrepository.interfaces.ScalarND;
import algorithmrepository.interfaces.VectorND;

public abstract class GeometryImpl implements Geometry {

    public static double[][][] leviCivita = null; {
         leviCivita = new double[3][3][3];
         
         leviCivita[0][1][2] = 1;
         leviCivita[1][2][0] = 1;
         leviCivita[2][0][1] = 1;
         
         leviCivita[0][2][1] = -1;
         leviCivita[2][1][0] = -1;
         leviCivita[1][0][2] = -1;
    }

    private static final double defaultStepLengthCoordDerivatives = 1e-8;
    double[] du;
    
    @Override
    public double[] coToContra(double[] atCoords, double[] coComponents) {
        int dim = coComponents.length;
        double[][] g = metricTensor(atCoords, Geometry.CONTRAVARIANT);
        
        double[] ret = new double[dim];
        for(int i=0;i<dim;++i) {
            for(int j=0;j<dim;++j) {
                ret[i] += g[i][j]*coComponents[j];
            }
        }
        
        return ret;        
    }
    
    @Override
    public double[][] metricTensor(double[] atCoords, int coOrContraMetric) {
        int dim = atCoords.length;
        double[][] g = new double[dim][dim];
        
        for(int i=0;i<dim;++i) {
            for(int j=i;j<dim;++j) {
                g[i][j] = metricCoefficient(atCoords, i, j, coOrContraMetric);
                g[j][i] = g[i][j];
            }
        }
        
        return g;
    }
    
    
    @Override
    public double[] contraToCo(double[] atCoords, double[] contraComponents) {
        int dim = atCoords.length;
        double[][] g = metricTensor(atCoords, Geometry.COVARIANT);
        
        double[] ret = new double[dim];
        for(int i=0;i<dim;++i) {
            for(int j=0;j<dim;++j) {
                ret[i] += g[i][j]*contraComponents[j];
            }
        }
        
        return ret;                
    }
    
    public double[][] metricTensorFromCoordinateTransform(double[] atCoords, int coOrContraVariant) {
        // d(x,y,z)/dui * d(x,y,z)/duj
        int dim = atCoords.length;
        
        double[] du = getStepLengthCoordDerivatives();
        double[][] dRdu = new double[dim][dim];

        for(int k=0;k<dim;++k) {
            double old = atCoords[k];
            atCoords[k] = old - du[k];
            double[] f0 = toCartesianCoords(atCoords);
            atCoords[k] = old + du[k];
            double[] f1 = toCartesianCoords(atCoords);
            dRdu[k][0] = Algorithms.der(f0[0], f1[0], du[k]);
            dRdu[k][1] = Algorithms.der(f0[1], f1[1], du[k]);
            dRdu[k][2] = Algorithms.der(f0[2], f1[2], du[k]);
            atCoords[k] = old;
        }
        
        double[][] g = new double[dim][dim];
        for(int i=0;i<dim;++i) {
            for(int j=0;j<dim;++j) {
                g[i][j] = dRdu[i][0]*dRdu[j][0] + dRdu[i][1]*dRdu[j][1] + dRdu[i][2]*dRdu[j][2];
            }
        }

        if (coOrContraVariant == Geometry.CONTRAVARIANT) {
            g = Algorithms.inv3x3(g);
        }
        
        return g;        
    }
        
    
    public double metricCoefficientFromCoordinateTransform(double[] atCoords, int i, int j, int coOrContraVariant) {
        // d(x,y,z)/dui * d(x,y,z)/duj
        int dim = atCoords.length;
        
        double[] du = getStepLengthCoordDerivatives();
               
        if (coOrContraVariant == COVARIANT) {
            double[] dRdui = new double[dim];
            double[] dRduj = new double[dim];

            double old = atCoords[i];
            atCoords[i] = old - du[i];
            double[] f0 = toCartesianCoords(atCoords);
            atCoords[i] = old + du[i];
            double[] f1 = toCartesianCoords(atCoords);
            dRdui[0] = Algorithms.der(f0[0], f1[0], du[i]);
            dRdui[1] = Algorithms.der(f0[1], f1[1], du[i]);
            dRdui[2] = Algorithms.der(f0[2], f1[2], du[i]);
            atCoords[i] = old;
            
            if (j != i) {
                old = atCoords[j];
                atCoords[j] = old - du[j];
                f0 = toCartesianCoords(atCoords);
                atCoords[j] = old + du[j];
                f1 = toCartesianCoords(atCoords);
                dRduj[0] = Algorithms.der(f0[0], f1[0], du[j]);
                dRduj[1] = Algorithms.der(f0[1], f1[1], du[j]);
                dRduj[2] = Algorithms.der(f0[2], f1[2], du[j]);
                atCoords[j] = old;
            } else {
                dRduj = dRdui.clone();
            }
               
            double ret = dRdui[0]*dRduj[0] + dRdui[1]*dRduj[1] + dRdui[2]*dRduj[2];
            
            return ret;        
        } else {
            double[][] dRdu = new double[dim][dim];

            for(int k=0;k<dim;++k) {
                double old = atCoords[k];
                atCoords[k] = old - du[k];
                double[] f0 = toCartesianCoords(atCoords);
                atCoords[k] = old + du[k];
                double[] f1 = toCartesianCoords(atCoords);
                dRdu[k][0] = Algorithms.der(f0[0], f1[0], du[k]);
                dRdu[k][1] = Algorithms.der(f0[1], f1[1], du[k]);
                dRdu[k][2] = Algorithms.der(f0[2], f1[2], du[k]);
                atCoords[k] = old;
            }
            
            // Jacobian
            double J = Algorithms.dot(dRdu[0], Algorithms.cross(dRdu[1], dRdu[2]));
            
            int[] cycl = cycl012(i);
            double[] gradui = Algorithms.cross(dRdu[cycl[0]], dRdu[cycl[1]]);
            for(int k=0;k<dim;++k) gradui[k] /= J;
            
            cycl = cycl012(j);
            double[] graduj = Algorithms.cross(dRdu[cycl[0]], dRdu[cycl[1]]);
            for(int k=0;k<dim;++k) graduj[k] /= J;
            
            double ret = gradui[0]*graduj[0] + gradui[1]*graduj[1] + gradui[2]*graduj[2]; 
            
            return ret;              
        }
    }
    
    private int[] cycl012(int i) {
        if (i==0) return new int[] { 1,2 };
        if (i==1) return new int[] { 2,0 };
        if (i==2) return new int[] { 0,1 };
        return null;
    }
    
  
    
    @Override
    public double[] flipCoContra(double[] atCoords, double[] components, int componentsCoContra) {
        if (componentsCoContra == Geometry.COVARIANT) return coToContra(atCoords, components);
        return contraToCo(atCoords, components);        
    }
    
    
    @Override
    public double dot(double[] atCoords, double[] v1Components, double[] v2Components, int coOrContra) {        
        int dim = atCoords.length;
        double[][] g = metricTensor(atCoords, coOrContra == COVARIANT ? CONTRAVARIANT : COVARIANT);
        double ret = 0;
        for(int i=0;i<dim;++i) {            
            for(int j=0;j<dim;++j) {
                ret += g[i][j]*v1Components[i]*v2Components[j];
            }
        }
        
        return ret;
    }

    @Override
    public double[] cross(double[] atCoords, double[] v1, double[] v2, int coOrContra) {
        int dim = atCoords.length;
        double[] ret = new double[dim];
        
        double sqrtg = jacobianDeterminant(atCoords);
        for(int i=0;i<dim;++i) {            
            for(int j=0;j<dim;++j) {
                ret[0] += leviCivita[i][j][0]/sqrtg*v1[i]*v2[j];
                ret[1] += leviCivita[i][j][1]/sqrtg*v1[i]*v2[j];
                ret[2] += leviCivita[i][j][2]/sqrtg*v1[i]*v2[j];                
            }
        }
         
        // ret comes out as co if vectors given in contra and the other way around.
        return flipCoContra(atCoords, ret, coOrContra == COVARIANT ? CONTRAVARIANT : COVARIANT);
    }
    
    
    @Override
    public double jacobianDeterminant(double[] atCoords) {        
        double[][] g = metricTensor(atCoords, Geometry.COVARIANT);
        
        if (g.length == 2) {
            return Algorithms.det2x2(g);
        } else if (g.length == 3) {
            return Algorithms.det3x3(g);
        } else {
            return determinant(g);
        }     
    }
    
    @Override
    public double[] jacobianDeterminant(double[][] atCoords) {
        int dim = atCoords.length;
        double[] ret = new double[atCoords[0].length];
        double[] pos = new double[dim];
        for(int i=0;i<ret.length;++i) {
            for(int j=0;j<dim;++j) pos[j] = atCoords[j][i];
            ret[i] = jacobianDeterminant(pos);
        }
        
        return ret;
    }
    
    
    /**
     * Following code taken from Dr Nicholas Duchon's page: http://sandsduchon.org/duchon/math/determinantJava.html
     * 
     * @param a The matrix whose determinant is sought.
     * @return The determinant of the matrix
     */
    private static double determinant(double [][] a) {
        int n = a.length - 1;
        if (n < 0) return 0;
        double M [][][] = new double [n+1][][];
        
        M[n] = a;  // init first, largest, M to a
        
        // create working arrays
        for (int i = 0; i < n; i++)
           M[i] = new double [i+1][i+1];
        
        return determinantRecursive(M, n);
     } 
     
     private static double determinantRecursive(double [][][] M, int m) {
       if (m == 0) return M[0][0][0];
       int e = 1;
       
       // init subarray to upper left mxm submatrix
       for (int i = 0; i < m; i++)
          for (int j = 0; j < m; j++)
             M[m-1][i][j] = M[m][i][j];
       double sum = M[m][m][m] * determinantRecursive (M, m-1);
       
       // walk through rest of rows of M
       for (int i = m-1; i >= 0; i--) {
         for (int j = 0; j < m; j++)
            M[m-1][i][j] = M[m][i+1][j];
         e = -e;
         sum += e * M[m][i][m] * determinantRecursive (M, m-1);
       } // end for each row of matrix
       
       return sum;
     } 
    
    @Override
    public double[] dl(double[][] atCoords, int forDim, double du) {
        int dim = atCoords.length;
        int numPoints = atCoords[0].length;
        double[] ret = new double[numPoints];
    
        double[] cur = new double[dim];
        for(int i=0;i<numPoints;++i) {    
            for(int j=0;j<dim;++j) cur[j] = atCoords[j][i];
            double[][] g = metricTensor(cur, Geometry.COVARIANT);
            ret[i] = Math.sqrt(g[forDim][forDim])*du;
        }
        
        return ret;
    }
    
    @Override
    public double[] dS(double[][] atCoords, int dim1, int dim2, double du1, double du2) {
        int dim = atCoords.length;
        int numPoints = atCoords[0].length;
        double[] ret = new double[numPoints];
    
        double[] cur = new double[dim];
        for(int i=0;i<numPoints;++i) {    
            for(int j=0;j<dim;++j) cur[j] = atCoords[j][i];
            double[][] g = metricTensor(cur, Geometry.COVARIANT);
            ret[i] = Math.sqrt(g[dim1][dim1]*g[dim2][dim2] - g[dim1][dim2]*g[dim1][dim2])*du1*du2;
        }
        
        return ret;
    }
    
    @Override
    public double[] dV(double[][] atCoords, double[] du) {
        int dim = atCoords.length;
        int numPoints = atCoords[0].length;
        double[] ret = new double[numPoints];
    
        double[] cur = new double[dim];
        for(int i=0;i<numPoints;++i) {    
            for(int j=0;j<dim;++j) cur[j] = atCoords[j][i];
            double sqrtg = jacobianDeterminant(cur);
            ret[i] = sqrtg;
            for(int j=0;j<dim;++j) ret[i] *= du[j];
        }
        
        return ret;
    }
    
    @Override
    public void setStepLengthCoordDerivatives(double[] du) {
        this.du = du.clone();        
    }
    
    @Override
    public double[] getStepLengthCoordDerivatives() {        
        if (du == null) {
             du = new double[dim()];
             for(int i=0;i<du.length;++i) du[i] = defaultStepLengthCoordDerivatives;
        }
        
        return du;
    }
    
    @Override
    public double christoffelSymbol2ndKind(double[] atCoords, int j, int i, int k) {
        int dim = atCoords.length;
        du = getStepLengthCoordDerivatives();
        
        double ret = 0;
        for(int n=0;n<dim;++n) {            
            ret += 0.5*metricCoefficient(atCoords, j, n, Geometry.CONTRAVARIANT)*(
                   metricCoeffDerivative(atCoords, n, i, k, Geometry.COVARIANT)+
                   metricCoeffDerivative(atCoords, n, k, i, Geometry.COVARIANT)-
                   metricCoeffDerivative(atCoords, i, k, n, Geometry.COVARIANT)
                   );
        }        
        
        return ret;
    }
    
    @Override
    public double christoffelSymbol1stKind(double[] atCoords, int j, int i, int k) {
        double ret = 0.5*(
                   metricCoeffDerivative(atCoords, j, i, k, Geometry.COVARIANT)+
                   metricCoeffDerivative(atCoords, j, k, i, Geometry.COVARIANT)-
                   metricCoeffDerivative(atCoords, i, k, j, Geometry.COVARIANT)
                   );
        
        return ret;
    }
    
    
    /**
     * Calculates dg(ij)/du(k), where g(ij) can be either the co or contra version of the metric tensor. 
     * 
     * 
     * @param atCoords
     * @param i Row index in metric tensor
     * @param j Column index in metric tensor
     * @param k Index of coordinate the derivative is with respect to
     * @param epsilon step size for numeric differentiation
     */
    public double metricCoeffDerivative(double[] atCoords, int i, int j, int k, int coOrContra) {        
        double mid = atCoords[k];
        double[] du = getStepLengthCoordDerivatives();
        
        atCoords[k] = mid-du[k];
        double f0 = metricCoefficient(atCoords, i, j, coOrContra);
        
        atCoords[k] = mid+du[k];
        double f1 = metricCoefficient(atCoords, i, j, coOrContra);
        
        atCoords[k] = mid;
        
        return Algorithms.der(f0, f1, du[k]);
    }
    
    
    @Override
    public double[] grad(ScalarND field, int returnCoOrContra, double[] atCoords) {
        int dim = atCoords.length;
        double[] ret = new double[dim];
        double[] du = getStepLengthCoordDerivatives();
        
        double[][] pos = new double[dim][1];
        for(int i=0;i<dim;++i) pos[i][0] = atCoords[i];
        for(int i=0;i<dim;++i) {
            
            double old = pos[i][0];
            
            pos[i][0] = old-du[i];
            double f0 = field.eval(pos)[0];
            pos[i][0] = old+du[i];
            double f1 = field.eval(pos)[0];            
            ret[i] = Algorithms.der(f0, f1, du[i]);
            
            pos[i][0] = old;            
        }        
        
        return returnCoOrContra == Geometry.COVARIANT ? ret : coToContra(atCoords, ret);
    }
    
    @Override
    public double[][] evalVectorField(final VectorND field, int fieldCoOrContra, int returnCoOrContra, double[][] atCoords) {
        int dim = atCoords.length;
        double[][] a = field.eval(atCoords);        
        
        if (fieldCoOrContra != returnCoOrContra) { // Flip component basis
            for(int i=0;i<a[0].length;++i) {
                double[] components = new double[dim];
                double[] coords = new double[dim];
                for(int j=0;j<dim;++j) components[j] = a[j][i];
                for(int j=0;j<dim;++j) coords[j] = atCoords[j][i];
                components = flipCoContra(coords, components, fieldCoOrContra);
                for(int j=0;j<dim;++j) a[j][i] = components[j];
            }
        }
        
        return a;        
    }
    
    @Override
    public double[] evalVectorField(final VectorND field, int fieldCoOrContra, int returnCoOrContra, double[] atCoords) {        
        int dim = atCoords.length;
        double[][] a = field.eval(singlePos(atCoords));        
        double[] components = new double[dim];
        for(int j=0;j<dim;++j) components[j] = a[j][0];
        
        if (fieldCoOrContra != returnCoOrContra) { // Flip component basis
            components = flipCoContra(atCoords, components, fieldCoOrContra);
        }
        
        return components;        
    }
    
    
    private static double[][] singlePos(double[] pos) {
        double[][] ret = new double[pos.length][1];
        for(int i=0;i<pos.length;++i) ret[i][0] = pos[i];
        return ret;
    }
    


    @Override
    public double div(VectorND field, int fieldCoOrContra, double[] atCoords) {
        int dim = atCoords.length;
        double[] du = getStepLengthCoordDerivatives();
        double sqrtg = jacobianDeterminant(atCoords);
        double ret = 0;
        
        double[][] pos = new double[dim][1];
        for(int i=0;i<dim;++i) pos[i][0] = atCoords[i];
        for(int i=0;i<dim;++i) {
            
            double old = pos[i][0];
            
            pos[i][0] = old-du[i];
            double f0 = evalVectorField(field, fieldCoOrContra, Geometry.CONTRAVARIANT, pos)[i][0]; 
            f0 *= jacobianDeterminant(pos)[0];
            pos[i][0] = old+du[i];
            double f1 = evalVectorField(field, fieldCoOrContra, Geometry.CONTRAVARIANT, pos)[i][0]; 
            f1 *= jacobianDeterminant(pos)[0];
                        
            ret += Algorithms.der(f0, f1, du[i]);
            
            pos[i][0] = old;                                    
        }
        ret *= 1.0/sqrtg;
        
        return ret;
    }
        
    /**
     * Calculates the derivative of a vector field component with respect to one coordinate, without taking into account the spatial change of basis, so
     * this is the derivative without the correction term that comes in through the Christoffer symbol. Don't know of a formal name for this part of the full
     * derivative.
     * 
     * @param field
     * @param fieldCoOrContra
     * @param returnCoOrContra
     * @param atCoords
     * @param du
     * @param dimComponent
     * @param dimDerivative
     * @return
     */
    private double vectorFieldSimpleDerivative(VectorND field, int fieldCoOrContra, int returnCoOrContra, double[] atCoords, double du, int dimComponent, int dimDerivative) {
        
        double old = atCoords[dimDerivative];        
        atCoords[dimDerivative] = old-du;
        double[] a = evalVectorField(field, fieldCoOrContra, Geometry.COVARIANT, atCoords);
        double f0 = a[dimComponent];
        atCoords[dimDerivative] = old+du;
        a = evalVectorField(field, fieldCoOrContra, Geometry.COVARIANT, atCoords);
        double f1 = a[dimComponent];
        atCoords[dimDerivative] = old;
        
        double der = Algorithms.der(f0, f1, du);
        
        return der;
    }
    
    @Override
    public double[] curl(VectorND field, int fieldCoOrContra, int returnCoOrContra, double[] atCoords) {
        int dim = atCoords.length; // == 3
        double[] du = getStepLengthCoordDerivatives();
        double sqrtg = jacobianDeterminant(atCoords);
        double[] ret = new double[dim];
                     
        ret[0] += vectorFieldSimpleDerivative(field, fieldCoOrContra, Geometry.COVARIANT, atCoords, du[1], 2, 1);
        ret[0] -= vectorFieldSimpleDerivative(field, fieldCoOrContra, Geometry.COVARIANT, atCoords, du[2], 1, 2);
        ret[0] *= 1.0/sqrtg;
        
        ret[1] += vectorFieldSimpleDerivative(field, fieldCoOrContra, Geometry.COVARIANT, atCoords, du[2], 0, 2);
        ret[1] -= vectorFieldSimpleDerivative(field, fieldCoOrContra, Geometry.COVARIANT, atCoords, du[0], 2, 0);
        ret[1] *= 1.0/sqrtg;
        
        ret[2] += vectorFieldSimpleDerivative(field, fieldCoOrContra, Geometry.COVARIANT, atCoords, du[0], 1, 0);
        ret[2] -= vectorFieldSimpleDerivative(field, fieldCoOrContra, Geometry.COVARIANT, atCoords, du[1], 0, 1);
        ret[2] *= 1.0/sqrtg;

        if (returnCoOrContra != Geometry.CONTRAVARIANT) {
            ret = flipCoContra(atCoords, ret, Geometry.CONTRAVARIANT);
        }
        
        return ret;
    }
    
    @Override
    public double[] fromCartesianCoords(double[] coordsXYZ) {
        throw new NotImplementedException();
    }
    
    
    @Override
    public double[][] cartesianBasisVectors(double[] atCoords, int coOrContraBasisVectors) {
        double[][] coBasis = transpose(jacobian(atCoords));
        
        if (coOrContraBasisVectors != Geometry.COVARIANT) {
            return flipCartesianBasis(atCoords, Geometry.COVARIANT, coBasis);
        } else {
            return coBasis;
        }
    }
    
    private final static double[][] transpose(double dIn[][]){
        int i,j,ni,nj;
        ni = dIn.length;
        if(ni <= 0)return new double[0][0];
        nj = dIn[0].length;
        double dOut[][] = new double[nj][ni];
        for(i=0;i<ni;i++)
            for(j=0;j<nj;j++)
                dOut[j][i]=dIn[i][j];
        
        return dOut;
        
    }
    
    @Override
    public double[][] jacobian(double[] atCoords) {
        int dim = atCoords.length;
        double[][] ret = new double[dim][dim];
        
        for(int i=0;i<dim;++i) {
            for(int j=0;j<dim;++j) {
                ret[i][j] = jacobianElement(atCoords, i, j);
            }
        }
        
        return ret;
    }
    
    @Override
    public double jacobianElement(double[] atCoords, int i, int j) {
        double[] du = getStepLengthCoordDerivatives();            
        double old = atCoords[j];
        
        atCoords[j] = old-du[j];
        double[] xyz = toCartesianCoords(atCoords);
        double f0 = xyz[i];
        
        atCoords[j] = old+du[j];
        xyz = toCartesianCoords(atCoords);
        double f1 = xyz[i];
        
        double ret  = Algorithms.der(f0, f1, du[j]);         
        
        atCoords[j] = old;
        
        return ret;
    }

     
     protected double[][] flipCartesianBasis(double[] atCoords, int coOrContra, double[][] cartesianBasis) {
         int dim = atCoords.length;
         double[][] g = metricTensor(atCoords, coOrContra == COVARIANT ? CONTRAVARIANT : COVARIANT);
         
         double[][] ret = new double[dim][dim];
         for(int i=0;i<dim;++i) {
             for(int j=0;j<dim;++j) {
                 for(int k=0;k<dim;++k) {
                     ret[i][j] += g[i][k]*cartesianBasis[k][j];
                 }                
             }
         }
         
         return ret;
     }

     @Override
     public double[] toCartesianComponents(double[] atCoords, double[] components, int componentsCoOrContra) {               
         double[][] baseInCartesian = cartesianBasisVectors(atCoords, componentsCoOrContra == COVARIANT ? CONTRAVARIANT : COVARIANT);
         
         int dim = components.length;
         double[] ret = new double[dim];
         for(int i=0;i<dim;++i) {
             for(int j=0;j<dim;++j) {
                 ret[i] += components[j]*baseInCartesian[j][i];
             }
         }
                 
         return ret;
     }

     @Override
     public double[] fromCartesianComponents(double[] atCoords, double[] componentsXYZ, int returnCoOrContra) {
         int dim = atCoords.length;

         if (dim != 3) {
             throw new RuntimeException("fromCartesianComponents only implemented for 3D, but dimension was "+dim);
         }

         double[][] e = returnCoOrContra == COVARIANT ? transpose(cartesianBasisVectors(atCoords, CONTRAVARIANT)) : transpose(cartesianBasisVectors(atCoords, COVARIANT));
         double[] components = mul(Algorithms.inv3x3(e), componentsXYZ);

         return components;
     }
     
     private static double[] mul(double[][] m, double[] v) {
         double[] ret = new double[v.length];
         
         for(int i=0;i<m.length;++i) {
             for(int j=0;j<m[i].length;++j) {
                 ret[i] += m[i][j]*v[j];
             }
         }
         
         return ret;
     }

 
}
